package ru.dbotthepony.kstarbound.client.gl.shader

import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.gl.GLType
import ru.dbotthepony.kstarbound.client.gl.vertex.GLAttributeList
import ru.dbotthepony.kstarbound.client.gl.vertex.StreamVertexBuilder
import ru.dbotthepony.kstarbound.client.gl.vertex.GeometryType
import ru.dbotthepony.kvector.arrays.Matrix4f
import ru.dbotthepony.kvector.vector.RGBAColor

private fun internalVertex(string: String) = StarboundClient.current().internalVertex(string)
private fun internalFragment(string: String) = StarboundClient.current().internalFragment(string)

private fun shaders(name: String): List<GLShader> {
	val client = StarboundClient.current()
	return listOf(client.internalVertex("shaders/$name.vsh"), client.internalFragment("shaders/$name.fsh"))
}

class GLLiquidProgram : GLShaderProgram(shaders("liquid"), FORMAT) {
	var baselineColor by F4Uniform("baselineColor")
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).build()
	}
}

class GLFlatColorProgram : GLShaderProgram(shaders("flat_color"), FORMAT) {
	var transform by F4x4Uniform("transform")

	val builder by lazy {
		StreamVertexBuilder(FORMAT, GeometryType.QUADS, 16384)
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC2F).push(GLType.VEC4F).build()
	}
}

class GLTileProgram : GLShaderProgram(shaders("tile"), FORMAT) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.identity()
		color = RGBAColor.WHITE
	}

	companion object {
		val FORMAT = GLAttributeList.Builder().push(GLType.VEC3F, GLType.VEC2F, GLType.FLOAT).build()
	}
}

class GLFontProgram : GLShaderProgram(shaders("font"), GLAttributeList.VERTEX_2D_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")
	var texture by IUniform("texture0")

	init {
		transform = Matrix4f.identity()
		color = RGBAColor.WHITE
	}
}

class GLFlatProgram : GLShaderProgram(shaders("flat"), GLAttributeList.VEC2F) {
	var transform by F4x4Uniform("transform")
	var color by F4Uniform("color")

	init {
		color = RGBAColor.WHITE
	}
}

class GLTexturedProgram : GLShaderProgram(listOf(internalVertex("shaders/vertex/texture.glsl"), internalFragment("shaders/fragment/texture.glsl")), GLAttributeList.VERTEX_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")

	val builder by lazy {
		StreamVertexBuilder(GLAttributeList.Builder().push(GLType.VEC3F, GLType.VEC2F).build(), GeometryType.QUADS, 16384)
	}

	init {
		transform = Matrix4f.identity()
	}
}

class GLTextured2dProgram : GLShaderProgram(listOf(internalVertex("shaders/vertex/2dtexture.glsl"), internalFragment("shaders/fragment/texture.glsl")), GLAttributeList.VERTEX_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")

	val builder by lazy {
		StreamVertexBuilder(GLAttributeList.Builder().push(GLType.VEC2F, GLType.VEC2F).build(), GeometryType.QUADS, 16384)
	}

	init {
		transform = Matrix4f.identity()
	}
}

class GLTexturedColoredProgram : GLShaderProgram(listOf(internalVertex("shaders/vertex/texture.glsl"), internalFragment("shaders/fragment/texture_color.glsl")), GLAttributeList.VERTEX_TEXTURE) {
	var transform by F4x4Uniform("transform")
	var texture by IUniform("texture0")
	var color by F4Uniform("color")

	init {
		transform = Matrix4f.identity()
		color = RGBAColor.WHITE
	}
}

class GLPrograms {
	val tile = GLTileProgram()
	val font = GLFontProgram()
	val flat = GLFlatProgram()
	val flatColor = GLFlatColorProgram()
	val liquid = GLLiquidProgram()
	val textured = GLTexturedProgram()
	val textured2d = GLTextured2dProgram()
	val texturedColored = GLTexturedColoredProgram()
}
