package ru.dbotthepony.kstarbound.client.render

import it.unimi.dsi.fastutil.longs.Long2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.longs.Long2ObjectOpenHashMap
import ru.dbotthepony.kvector.arrays.Matrix4fStack

/**
 * Позволяет вызывать отрисовщики в определённой (послойной) последовательности
 */
class LayeredRenderer {
	private val layers = Long2ObjectAVLTreeMap<ArrayList<(Matrix4fStack) -> Unit>>()
	private val layersHash = Long2ObjectOpenHashMap<ArrayList<(Matrix4fStack) -> Unit>>()

	fun add(layer: Long, renderer: (Matrix4fStack) -> Unit) {
		var list = layersHash[layer]

		if (list == null) {
			list = ArrayList()
			layers[layer] = list
			layersHash[layer] = list
		}

		list.add(renderer)
	}

	fun render(stack: Matrix4fStack) {
		for (list in layers.values) {
			for (renderer in list) {
				renderer.invoke(stack)
			}
		}
	}
}
